import unittest

from Testing.ZopeTestCase import ZopeTestCase

from . import common


deprecations = {
    "z3c.jbot.tests.templates.example_deprecated.pt": "z3c.jbot.tests.templates.example.pt"  # noqa: E501
}


class FiveTests(ZopeTestCase):
    def setUp(self):
        common.setUp(self)
        super().setUp()

        from Products.Five.browser import BrowserView
        from Products.Five.browser.pagetemplatefile import \
            ZopeTwoPageTemplateFile as Template

        class MockView(BrowserView):
            template = Template("templates/example.pt")
            interface_override = Template(
                "overrides/interface/z3c.jbot.tests.templates.example.pt"
            )
            interface_deprecated_override = Template(
                "overrides/interface_deprecated/z3c.jbot.tests.templates.example_deprecated.pt"  # noqa: E501
            )
            http_override = Template(
                "overrides/http/z3c.jbot.tests.templates.example.pt"
            )
            https_override = Template(
                "overrides/https/z3c.jbot.tests.templates.example.pt"
            )

        # set up mock site and request
        from zope import component
        from zope.publisher.browser import TestRequest

        class MockSite:
            REQUEST = TestRequest("en")
            getSiteManager = component.getSiteManager

        from zope.component.hooks import setHooks
        from zope.component.hooks import setSite

        setHooks()
        setSite(MockSite())
        self._request = MockSite.REQUEST

        # render templates for later comparison
        view = self._view = MockView(self.folder, MockSite.REQUEST)
        self._original = view.template()
        self._interface_override = view.interface_override()
        self._interface_deprecated_override = view.interface_deprecated_override()  # noqa: E501
        self._http_override = view.http_override()
        self._https_override = view.https_override()

        import z3c.jbot.tests

        self._tests = z3c.jbot.tests.__path__[0]

    def tearDown(self):
        import zope.component.testing

        zope.component.testing.tearDown(self)

    def test_override_for_interface(self):
        from zope import interface

        from z3c.jbot.metaconfigure import handler

        handler("%s/overrides/interface" % self._tests, interface.Interface)
        self.assertEqual(self._view.template(), self._interface_override)

    def test_override_for_httprequest(self):
        from zope import interface
        from zope.publisher.interfaces.browser import IHTTPRequest

        from z3c.jbot.metaconfigure import handler

        class IHTTPSRequest(IHTTPRequest):
            pass

        # register handlers
        handler("%s/overrides/interface" % self._tests, interface.Interface)
        handler("%s/overrides/https" % self._tests, IHTTPSRequest)

        # we get the general override
        self.assertEqual(self._view.template(), self._interface_override)

        # provide the http layer
        from zope.interface import alsoProvides

        alsoProvides(self._request, IHTTPSRequest)
        self.assertEqual(self._view.template(), self._https_override)

        # revert to general override
        from zope.interface import noLongerProvides

        noLongerProvides(self._request, IHTTPSRequest)
        self.assertEqual(self._view.template(), self._interface_override)

    def test_zcml_features(self):
        from zope.configuration import xmlconfig
        from zope.configuration.config import ConfigurationMachine

        context = ConfigurationMachine()
        xmlconfig.registerCommonDirectives(context)

        xmlconfig.string(
            """
            <configure
                xmlns="http://namespaces.zope.org/zope"
            >
                <include package="z3c.jbot" file="meta.zcml" />
            </configure>
            """, context=context
        )

        self.assertTrue(context.hasFeature("jbot-deprecations"))

    def test_deprecated_override(self):
        from unittest.mock import patch

        from zope import interface

        from z3c.jbot.metaconfigure import handler

        overrides_folder = f"{self._tests}/overrides/interface_deprecated"

        # Without the deprecation in place we do not get any warnings
        with patch("warnings.warn") as mock_warn:
            handler(overrides_folder, interface.Interface)
            self.assertEqual(mock_warn.call_count, 0)

        # and the view still uses the original template because the override
        # is referring to a not existent template
        self.assertEqual(self._view.template(), self._original)

        # We now set up the deprecation
        from zope.configuration import xmlconfig
        xmlconfig.string(
            """
            <configure
                xmlns="http://namespaces.zope.org/zope"
                xmlns:browser="http://namespaces.zope.org/browser"
            >
                <include package="z3c.jbot" file="meta.zcml" />
                <browser:jbotDeprecated dictionary="z3c.jbot.tests.test_five.deprecations" />
            </configure>
            """  # noqa: E501
        )

        # And rerun the handler
        with patch("warnings.warn") as mock_warn:
            handler(
                overrides_folder, interface.Interface
            )
            self.assertEqual(mock_warn.call_count, 1)
            self.assertTupleEqual(
                mock_warn.call_args.args,
                (
                    f'Template {self._tests}/overrides/interface_deprecated/z3c.jbot.tests.templates.example_deprecated.pt '  # noqa: E501
                    f'deprecated, use z3c.jbot.tests.templates.example.pt instead.',  # noqa: E501
                )
            )

        # The view now is able to resolve the renamed template
        self.assertEqual(
            self._view.template(),
            self._interface_deprecated_override,
        )


def test_suite():
    return unittest.TestSuite(
        (unittest.defaultTestLoader.loadTestsFromTestCase(FiveTests),))



import logging

from zope.pagetemplate.pagetemplatefile import PageTemplateFile

from . import utility


try:
    from Acquisition.interfaces import IAcquirer
except ModuleNotFoundError:
    IAcquirer = None


logger = logging.getLogger('jbot')

# Standard PageTemplateFile

PT_CLASSES = [PageTemplateFile]

try:
    import Products.PageTemplates.PageTemplateFile
    PT_CLASSES.append(Products.PageTemplates.PageTemplateFile.PageTemplateFile)
except ModuleNotFoundError:
    pass

registry = {}


def get(template, view=None, cls=None):
    layer = utility.getLayer()
    key = layer, template
    inst = registry.get(key)
    if inst is None:
        cls = type(template)
        inst = registry[key] = cls.__new__(cls)
        inst.__dict__ = template.__dict__.copy()

    for manager in utility.getManagers(layer):
        # register template; this call returns ``True`` if the
        # template was invalidated (changed filename)
        if manager.registerTemplate(inst, template):
            inst._v_last_read = False
            inst.__dict__.pop('_v_template', None)
            break

    if view is not None and IAcquirer is not None:
        if IAcquirer.providedBy(inst) and IAcquirer.providedBy(view):
            return inst.__of__(view)

    return inst


# five.pt / Chameleon
try:
    from five.pt.pagetemplate import ViewPageTemplateFile as pt_class
except ModuleNotFoundError:
    pass
else:
    five_bind = pt_class.__get__

    def get_and_bind(template, view=None, cls=None):
        inst = get(template, view, cls)
        if inst._v_last_read is False:
            inst.registry.purge()
            inst.read()
        return five_bind(inst, view, cls)

    pt_class.__get__ = get_and_bind
    logger.debug(repr(pt_class))

    del pt_class

# Zope 2.12 ViewPageTemplateFile; note that we import
# ``BoundPageTemplate`` to provoke an import-error on Zope 2.10.
try:
    from Products.Five.browser.pagetemplatefile import \
        ViewPageTemplateFile as pt_class
    zope_bind = pt_class.__get__
except ModuleNotFoundError:
    pass
except AttributeError:
    pass
else:
    def five_get_and_bind(template, view=None, cls=None):
        inst = get(template, view, cls)
        if inst._v_last_read is False:
            inst.read()
        return zope_bind(inst, view, cls)

    pt_class.__get__ = five_get_and_bind
    logger.debug(repr(pt_class))

    del pt_class

# zope.browserpage ViewPageTemplateFile
try:
    from zope.browserpage.viewpagetemplatefile import \
        ViewPageTemplateFile as pt_class
    browserpage_bind = pt_class.__get__
except ModuleNotFoundError:
    pass
except AttributeError:
    pass
else:
    def get_and_bind(template, view=None, cls=None):
        inst = get(template, view, cls)
        if inst._v_last_read is False:
            inst.read()
        return browserpage_bind(inst, view, cls)

    pt_class.__get__ = get_and_bind
    logger.debug(repr(pt_class))

    del pt_class


for pt_class in PT_CLASSES:
    pt_class.__get__ = get
    logger.debug(repr(pt_class))

# CMF skin layer resources
try:
    from Products.CMFCore.FSObject import FSObject as fs_class
except ModuleNotFoundError:
    pass
else:
    of = fs_class.__of__

    def get_skin_obj(obj, view=None, cls=None):
        layer = utility.getLayer()
        key = layer, obj
        inst = registry.get(key)
        if inst is None:
            cls = obj.__class__
            inst = registry[key] = cls.__new__(cls)
            inst.__dict__ = obj.__dict__.copy()

        for manager in utility.getManagers(layer):
            # register template; this call returns ``True`` if the
            # template was invalidated (changed filename)
            if manager.registerTemplate(inst, obj):
                inst._parsed = False
                inst.getObjectFSPath()

        return of(inst, view)

    def get_filename(obj, *args):
        return obj._filepath

    def set_filename(obj, value, *args):
        obj._filepath = value

    fs_class.__of__ = get_skin_obj
    fs_class.filename = property(get_filename, set_filename)

    logger.debug(repr(fs_class))



import doctest
import unittest

import zope.component
import zope.interface

from .common import setUp
from .common import tearDown


OPTIONFLAGS = (
    doctest.ELLIPSIS | doctest.NORMALIZE_WHITESPACE
)


def test_suite():
    globs = dict(
        interface=zope.interface,
        component=zope.component)

    return unittest.TestSuite((
        doctest.DocFileSuite(
            'README.rst',
            optionflags=OPTIONFLAGS,
            setUp=setUp,
            tearDown=tearDown,
            globs=globs,
            package="z3c.jbot",
        ),
    ))



import shutil
import tempfile

import zope.component.testing
import zope.configuration.xmlconfig


def setUp(test):
    zope.component.testing.setUp(test)

    import z3c.jbot
    zope.configuration.xmlconfig.XMLConfig('configure.zcml', z3c.jbot)()

    # enable five.pt if present
    try:
        import five.pt
    except ModuleNotFoundError:
        pass
    else:
        zope.configuration.xmlconfig.XMLConfig('configure.zcml', five.pt)()

    test.tempdir = tempfile.tempdir
    tempfile.tempdir = tempfile.mkdtemp()


def tearDown(test):
    zope.component.testing.tearDown(test)

    try:
        tempdir = test.tempdir
    except AttributeError:
        pass
    else:
        try:
            shutil.rmtree(tempfile.tempdir)
        finally:
            tempfile.tempdir = tempdir



from zope.configuration import fields
from zope.configuration.fields import GlobalObject
from zope.interface import Interface


class ITemplateOverridesDirective(Interface):
    """Directive which registers a directory with template overrides."""

    directory = fields.Path(
        title="Path to directory",
        required=True,
    )

    layer = GlobalObject(
        title="The layer the overrides should be enabled for",
        description="By default overrides are used for all layers.",
        required=False,
    )


class IDeprecatedTemplatesDirective(Interface):
    """Directive which registers a dictionary with deprecated templates.

    This can be used when you move templates and want an existing jbot
    override for the old template path to still work.

    Example:

    * You move old_package/test.pt to new_package/test.pt.
    * You register a dictionary with:
      {"old_package.test.pt": "new_package.test.pt"}
    * If a third party package has a template override old_package.test.pt,
      we display a warning that the package should use new_package.test.pt.
    * The override with the old name still works:
      instead of new_package/test.pt, the override is used.
    * Of course an override with the new name works as well.
    """

    dictionary = GlobalObject(
        title="Python path to dictionary",
        required=True,
    )



from zope.interface import implementer
from zope.interface import providedBy
from zope.publisher.interfaces.browser import IBrowserPublisher


try:
    from plone.resource.file import FilesystemFile
    HAS_PLONE_RESOURCE = True
except ModuleNotFoundError:
    HAS_PLONE_RESOURCE = False

from .utility import getManagers


@implementer(IBrowserPublisher)
class FilesystemFileResourceBrowserPublisher:
    def __init__(self, context, request):
        self.context = context
        self.request = request

    def browserDefault(self, request):
        assert HAS_PLONE_RESOURCE
        layer = providedBy(request)
        for manager in getManagers(layer):
            path = manager.queryResourcePath(self.context)
            if path is not None:
                resource = FilesystemFile(
                    self.context.__parent__,
                    request,
                    path,
                    self.context.__name__,
                )
                break
        else:
            resource = self.context

        return resource, ()



import os
import sys
import warnings

from zope.interface import implementer

from . import interfaces
from . import utility


IGNORE = object()
DELETE = object()
# Dictionary of templates that have been moved to a new location,
# and for which we want an override for the old location to still work
# for the new location as well.
DEPRECATED_TEMPLATES_DICT = {}


def update_deprecated_templates_dict(dictionary):
    DEPRECATED_TEMPLATES_DICT.update(dictionary)


def normalize(filepath):
    return os.path.normcase(os.path.normpath(filepath))


def root_length(a, b):
    if b.startswith(a):
        return len(a)
    return 0


def sort_by_path(path, paths):
    return sorted(
        paths, key=lambda syspath: root_length(syspath, path), reverse=True
    )


def find_zope2_product(path):
    """Check the Zope 5 magic Products semi-namespace to see if the
    path is part of a Product."""
    _syspaths = sort_by_path(
        path,
        [normalize(path) for path in sys.modules["Products"].__path__],
    )
    syspath = _syspaths[0]

    if not path.startswith(syspath):
        return

    product = path[len(syspath) + 1:].split(os.path.sep, 2)[0]

    return normalize("Products." + product)


def find_package(syspaths, path):
    """Determine the Python-package where path is located.  If the path is
    not located within the Python sys-path, return ``None``.
    The returned path is already 'normcase'. """

    _syspaths = sort_by_path(path, syspaths)
    syspath = _syspaths[0]

    if not path.startswith(syspath):
        if not utility.ZOPE_3:
            return find_zope2_product(path)
        return None

    path = path[len(syspath):]

    # convert path to dotted filename
    if path.startswith(os.path.sep):
        path = path[1:]
    return path


class ResourceManagerFactory:
    def __init__(self, name):
        self.manager = TemplateManager(name)

    def __call__(self, layer):
        return self.manager


class TemplateManagerFactory:
    def __init__(self, name):
        self.manager = TemplateManager(name)

    def __call__(self, layer):
        return self.manager


@implementer(interfaces.ITemplateManager)
class TemplateManager:
    def __init__(self, name):
        self.syspaths = {normalize(p) for p in sys.path}
        self.resources = {}
        self.templates = {}
        self.paths = {}
        self.directories = set()
        self.name = name

    def registerDirectory(self, directory):
        directory = normalize(directory)
        self.directories.add(directory)

        for filename in os.listdir(directory):
            filename = os.path.normcase(filename)
            full_path = normalize(f"{directory}{os.path.sep}{filename}")
            self.paths[filename] = full_path
            canonical_filename = DEPRECATED_TEMPLATES_DICT.get(filename)
            if canonical_filename:
                warnings.warn(
                    f"Template {full_path} deprecated, "
                    f"use {canonical_filename} instead."
                )
                self.paths[canonical_filename] = full_path

        for template, filename in list(self.templates.items()):
            if filename is IGNORE:
                del self.templates[template]

    def unregisterDirectory(self, directory):
        directory = normalize(directory)
        self.directories.remove(directory)

        templates = []

        for template, filename in self.templates.items():
            if os.path.normcase(filename) in self.paths:
                templates.append(template)

        for filename in os.listdir(directory):
            if os.path.normcase(filename) in self.paths:
                del self.paths[filename]
        for template in templates:
            inst = template.__get__(self)
            self.registerTemplate(inst, template)
            del self.templates[template]
            inst.filename = inst._filename

    def unregisterAllDirectories(self):
        for directory in tuple(self.directories):
            self.unregisterDirectory(directory)

    def registerTemplate(self, template, token):
        # assert that the template is not already registered
        filename = self.templates.get(token)
        if filename is IGNORE:
            return

        # if the template filename matches an override, we're done
        if filename is not None:
            if self.paths.get(filename) == template.filename:
                return

        # verify that override has not been unregistered
        if filename is not None and filename not in self.paths:
            template.filename = template._filename
            del self.templates[token]

        # check if an override exists
        path = find_package(self.syspaths, normalize(template.filename))
        if path is None:
            # permanently ignore template
            self.templates[token] = IGNORE
            return

        filename = path.replace(os.path.sep, '.')
        if filename not in self.paths:
            self.templates[token] = IGNORE
            template._v_last_read = False
            return

        path = self.paths[filename]

        # save original filename
        template._filename = template.filename

        # save template and registry and assign path
        template.filename = path
        self.templates[token] = os.path.normcase(filename)

        return True

    def queryResourcePath(self, resource):
        path = self.resources.get(resource.path)
        if path is IGNORE:
            return

        if path is not None:
            return path

        path = find_package(self.syspaths, normalize(resource.path))
        if path is None:
            self.resources[resource.path] = IGNORE
            return

        filename = path.replace(os.path.sep, '.')
        resource_path = self.paths.get(filename)
        if resource_path is None:
            self.resources[resource.path] = IGNORE
            return

        return resource_path



from zope import component
from zope import interface
from zope.publisher.interfaces.browser import IBrowserPublisher

from .manager import update_deprecated_templates_dict


try:
    from plone.resource.file import FilesystemFile
    HAS_PLONE_RESOURCE = True
except ModuleNotFoundError:
    HAS_PLONE_RESOURCE = False

from . import browser
from . import interfaces
from . import manager


def handler(directory, layer):
    lookup_all = component.getGlobalSiteManager().adapters.lookupAll

    # check if a template manager already exists
    factories = {
        factory
        for name, factory in lookup_all((layer,), interfaces.ITemplateManager)
    }

    # this might yield several factories (template managers); we check
    # if one is registered for exactly our layer
    base_factories = set()
    if layer is not interface.Interface:
        for base in layer.__bases__:
            for name, factory in lookup_all(
                (base,), interfaces.ITemplateManager
            ):
                base_factories.add(factory)

    try:
        factory = factories.difference(base_factories).pop()
    except KeyError:
        name = directory
        factory = manager.TemplateManagerFactory(directory)
        component.provideAdapter(
            factory, (layer,), interfaces.ITemplateManager, name=name
        )

        if HAS_PLONE_RESOURCE:
            component.provideAdapter(
                browser.FilesystemFileResourceBrowserPublisher,
                (FilesystemFile, layer), IBrowserPublisher
            )

    template_manager = factory(layer)
    template_manager.registerDirectory(directory)

    return template_manager


def templateOverridesDirective(_context, directory, layer=interface.Interface):
    # All template overrides must be loaded *after* the deprecations from
    # the deprecatedTemplatesDirective have been loaded, otherwise the
    # deprecations are not effective.  So set a higher order here.
    # zope.configuration uses the 'order' key for determining the
    # order in which it executes any registered actions.
    _context.action(
        discriminator=('jbot', directory, layer),
        callable=handler,
        args=(directory, layer),
        order=1000,
    )


def deprecatedTemplatesDirective(_context, dictionary):
    # Discriminators must be hashable.
    dict_discriminator = "-".join(sorted(dictionary.keys()))
    _context.action(
        discriminator=('jbot', dict_discriminator),
        callable=update_deprecated_templates_dict,
        args=(dictionary,),
        # Explicitly set the order just to be clear, although zero is the
        # default.
        order=0,
    )



from setuptools import find_packages
from setuptools import setup


setup(
    name='z3c.jbot',
    version='2.2',
    description="Drop-in template overrides.",
    long_description=(open('README.rst').read() + "\n" +
                      open('CHANGES.rst').read()),
    long_description_content_type='text/x-rst',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Framework :: Zope',
        'Framework :: Zope :: 3',
        'Framework :: Zope :: 5',
        'License :: OSI Approved :: Zope Public License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Programming Language :: Python :: 3.12',
        'Programming Language :: Python :: 3.13',
    ],
    keywords='page template override',
    author='Zope Foundation and Contributors',
    author_email='zope-dev@zope.dev',
    url='https://github.com/zopefoundation/z3c.jbot',
    license='ZPL-2.1',
    packages=find_packages('src'),
    package_dir={'': 'src'},
    namespace_packages=['z3c'],
    include_package_data=True,
    zip_safe=False,
    python_requires='>=3.9',
    install_requires=[
        'setuptools',
        'zope.pagetemplate',
        'zope.component',
        'zope.configuration',
        'zope.security',
        'zope.publisher',
    ],
    extras_require={
        'test': [
            'Zope',
            'Products.BTreeFolder2',
            'Products.CMFCore',
            'plone.resource',
        ],
    },
    entry_points="""
    [z3c.autoinclude.plugin]
    target = plone
    """,
)



import zope.security.interfaces
import zope.security.management
from zope.component import getGlobalSiteManager
from zope.component.hooks import getSite
from zope.interface import Interface
from zope.interface import providedBy
from zope.publisher.interfaces import IRequest

from z3c.jbot.interfaces import ITemplateManager


try:
    import Acquisition  # noqa
    ZOPE_3 = False
except ModuleNotFoundError:
    ZOPE_3 = True


def getRequest():
    if not ZOPE_3:
        # get request by acquisition
        site = getSite()
        if site is not None:
            try:
                return site.request
            except AttributeError:
                return getattr(site, "REQUEST", None)

    try:
        i = zope.security.management.getInteraction()
    except zope.security.interfaces.NoInteraction:
        return

    for p in i.participations:
        if IRequest.providedBy(p):
            return p


def getLayer():
    request = getRequest()

    if request is not None:
        return providedBy(request)

    return Interface


def getManagers(layer):
    try:
        adapters = getGlobalSiteManager().adapters._adapters[1]
    except IndexError:
        return

    for iface in layer.__sro__:
        by_interface = adapters.get(iface)

        if by_interface is not None:
            managers = by_interface.get(ITemplateManager)

            if managers is not None:
                items = managers.items()
                if len(items) > 1:
                    items = sorted(items)

                for name, factory in items:
                    yield factory(layer)
